#!/usr/bin/env python
"""
Standard plotting class of Matthias Cuntz.

It has the same functionality as the old mc_template.py by Matthias Cuntz but
uses the object-oriented approach of st_template.py of Stephan Thober.

It allows plotting on screen, into PDF and PNG files, as well as in HTML
files as a wrapper for PNG images or using hvplot.

It is optimised for publication ready plots, either on white or black
background.


The simplest way to use mcPlot is to extend the class:

.. code-block:: python

   import numpy as np
   from pyjams import mcPlot

   class PlotIt(mcPlot):
       def __init__(self, *args, **kwargs):
           super().__init__(*args, **kwargs)
           # change e.g. colors
           self.lcol1 = 'cyan'
           # reset global values after colour changes, etc.
           self.set_matplotlib_rcparams()

       def read_data(self):
           # do something
           self.dat = np.arange(10)

       def plot_fig_1(self):
           import matplotlib.pyplot as plt

           self.ifig += 1
           fig = plt.figure(self.ifig)

           sub = fig.add_axes([0.125, 0.667, 0.3375, 0.233])

           larr = sub.plot(self.dat)
           plt.setp(larr[-1], linestyle='-', linewidth=self.lwidth,
                    marker='', color=self.lcol1)

           self.plot_save(fig)

       def plot_fig_2(self):
           import matplotlib.pyplot as plt

           self.ifig += 1
           fig = plt.figure(self.ifig)

           sub = fig.add_axes([0.125, 0.667, 0.3375, 0.233])

           larr = sub.plot(2*self.dat)
           plt.setp(larr[-1], linestyle='-', linewidth=self.lwidth,
                    marker='', color=self.lcols[-1])

           self.plot_save(fig)

   if __name__ == '__main__':
       iplot = PlotIt(desc='Test mcPlot',
                      argstr='No argument wanted')
       iplot.read_data()
       iplot.plot_fig_1()
       iplot.plot_fig_2()
       iplot.close()

Then call the script with -h to see the command line options.


This module was written by Matthias Cuntz while at Institut National de
Recherche pour l'Agriculture, l'Alimentation et l'Environnement (INRAE), Nancy,
France.

:copyright: Copyright 2020-2022 Matthias Cuntz, see AUTHORS.rst for details.
:license: MIT License, see LICENSE for details.

.. moduleauthor:: Matthias Cuntz

The following classes are provided:

.. autosummary::
   mcPlot

History
    * Written Sep 2020 by Matthias Cuntz (mc (at) macu (dot) de)
    * Write standard output file into current folder, Nov 2021, Matthias Cuntz
    * Change from NCL amwg color palette to pyjams amwg,
      May 2021, Matthias Cuntz
    * Add **kwargs to plot_save, May 2022, Matthias Cuntz
    * Add self.transparent to pdf output in plot_save, May 2022, Matthias Cuntz
    * Add --transparent as a standard option, May 2022, Matthias Cuntz
    * Add left, bottom, top to standard layout options,
      Jul 2022, Matthias Cuntz
    * Add --dpi as a standard option, Jan 2023, Matthias Cuntz
    * Use helper.filebase, Mar 2023, Matthias Cuntz
    * Replace plotly with hvplot, May 2023, Matthias Cuntz
    * Set filename without suffix as default plot name,
      Jun 2023, Matthias Cuntz
    * Removed space from print of plot filename, Jun 2023, Matthias Cuntz

"""
import numpy as np
from .helper import filebase


__all__ = ['mcPlot']


# -------------------------------------------------------------------------
# Class mcPlot
#

class mcPlot(object):
    """
    Standard plotting class of Matthias Cuntz

    Upon initialisation,
    the command line arguments are gathered (get_command_line_arguments),
    the output type is set (set_output_type),
    standard layout options are added to self (set_layout_options),
    global rcParams are set for Matplotlib (set_matplotlib_rcparams), and
    the output plotting file is opened (plot_begin).

    Attributes
    ----------
    desc : string, optional
        Description for command line parser, which will be shown when
        called with -h.
    argstr : string, optional
        String given as description for the positional arguments.

    Methods
    -------
    get_command_line_arguments(desc=None, argstr=None)
        Standard command line parser with the default arguments
        such as plot type, filename, etc. If extra arguments are needed,
        one should copy this routine into the extending class and adapt
        it to its needs, keeping the existing optional arguments.
    plot_end() or plot_stop() or plot_close() or end() or stop()
        Finish, closing opened output files.
    plot_save(fig, **kwargs)
        Save, close or show `figure`.
    set_layout_options()
        Sets the colours and styles that can be used in plots.
        One can either copy this routine into the extending class and adapt it,
        or add a new method that resets some of the layout options and call it
        in the initialisation of the extending class, or simply set layout
        options in the initialisation of the extending class.

    Notes
    -----
    Several more methods are defined, which should probably not be changed.

    plot_begin() or plot_start()
        Open output file and similar at the beginning.

    plot_test()
        A simple plot as an example.

    set_matplotlib_rcparams()
        Set rcParams of Matplotlib depending on output type, and chosen layout.
        rcParams can also be re-set in the initialisation of the extending
        class.

    set_output_type()
        Set the format of the output such as pdf or png.

    Examples
    --------
    The simplest way to use `mcPlot` is to extend the class:

    .. code-block:: python

       import numpy as np
       from pyjams import mcPlot

       class PlotIt(mcPlot):
           def __init__(self, *args, **kwargs):
               super().__init__(*args, **kwargs)
               # change e.g. colors
               self.lcol1 = 'cyan'

           def read_data(self):
               # do something
               self.dat = np.arange(10)

           def plot_fig_1(self):
               import matplotlib.pyplot as plt

               self.ifig += 1
               fig = plt.figure(self.ifig)

               sub = fig.add_axes([0.125, 0.667, 0.3375, 0.233])

               larr = sub.plot(self.dat)
               plt.setp(larr[-1], linestyle='-', linewidth=self.lwidth,
                        marker='', color=self.lcol1)

               self.plot_save(fig)

           def plot_fig_2(self):
               import matplotlib.pyplot as plt

               self.ifig += 1
               fig = plt.figure(self.ifig)

               sub = fig.add_axes([0.125, 0.667, 0.3375, 0.233])

               larr = sub.plot(2*self.dat)
               plt.setp(larr[-1], linestyle='-', linewidth=self.lwidth,
                        marker='', color=self.lcols[-1])

               self.plot_save(fig)

       if __name__ == '__main__':
           iplot = PlotIt(desc='Test mcPlot')
           iplot.read_data()
           iplot.plot_fig_1()
           iplot.plot_fig_2()
           iplot.close()

    Then call the script with -h to see the command line options.

    """

    # -------------------------------------------------------------------------
    # init
    #
    def __init__(self, desc=None, argstr=None):
        """
        Initialise the class mcPlot.

        It does the following steps:
            the command line arguments are gathered,
            the output type is set,
            standard layout options are added to self,
            global rcParams are set for Matplotlib, and
            the output plotting file is opened.

        Examples
        --------
        An extending class should initialise with something similar to

        .. code-block:: python
           class UsemcPlot(mcPlot):
               def __init__(self, *args, **kwargs):
                   super().__init__(*args, **kwargs)`

        """
        # get options
        self.get_command_line_arguments(desc=desc, argstr=argstr)
        # pdf, png, ...
        self.set_output_type()
        # nrow, ncol, colours, etc.
        self.set_layout_options()
        # mpl.use and rcParams
        self.set_matplotlib_rcparams()
        # begin plot
        self.plot_begin()

    # -------------------------------------------------------------------------
    # command line arguments
    #
    def get_command_line_arguments(self, desc=None, argstr=None):
        """
        Standard command line parser with default arguments
        such as plot type, filename, etc.

        If extra arguments are needed, one should copy this routine
        into an extending class and adapt it to its needs,
        keeping the existing optional arguments.

        Parameters
        ----------
        desc : string, optional
            Description for command line parser, which will be shown when
            called with -h.
        argstr : string, optional
            String given as description for the positional arguments.

        Notes
        -----
        Standard command line arguments are:

        .. code-block:: bash

           positional arguments:
             args                  Text will be replaced by `argstr`.

           optional arguments:
             -h, --help            show this help message and exit
             -p plotname, --plotname plotname
                                   Name of plot output file for types pdf,
                                   html, d3, or hvplot, and name basis for type
                                   png (default: calling_filename without
                                   extension).
             -s, --serif           Use serif font; default sans serif.
             -t outtype, --type outtype
                                   Output type is pdf, png, html, d3, or hvplot
                                   (default: open screen windows).
             -u, --usetex          Use LaTeX to render text in pdf, png and
                                   html.
             -w, --white           White lines on transparent or black
                                   background; default: black lines on
                                   transparent or white background.
             --dpi number          Dots Per inch (DPI) for non-vector output
                                   types or rasterized maps in vector output
                                   (default: 300).
             --transparent         Transparent figure background
                                   (default: black or white).

        Examples
        --------
        .. code-block:: python

           iplot = mcPlot(desc="Test Matthias' plotting class.",
                          argstr="directory file")

        """
        import argparse
        import os

        if desc is None:
            idesc = "Matthias Cuntz' standard plotting class."
        else:
            idesc = desc
        if argstr is None:
            iargstr = 'Command line arguments.'
        else:
            iargstr = argstr

        plotname = filebase(os.path.basename(__file__))
        serif    = False
        outtype  = ''
        transparent = False
        usetex   = False
        dowhite  = False
        dpi      = 300
        parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=idesc)
        hstr = (f'Name of plot output file for types pdf, html, d3, or'
                f' hvplot, and name basis for type png (default:'
                f' {plotname}).')
        parser.add_argument('-p', '--plotname', action='store',
                            default=plotname, dest='plotname',
                            metavar='plotname', help=hstr)
        hstr = 'Use serif font; default sans serif.'
        parser.add_argument('-s', '--serif', action='store_true',
                            default=serif, dest='serif', help=hstr)
        hstr = ('Output type is pdf, png, html, d3, or hvplot'
                ' (default: open screen windows).')
        parser.add_argument('-t', '--type', action='store', default=outtype,
                            dest='outtype', metavar='outtype', help=hstr)
        hstr = 'Use LaTeX to render text in pdf, png and html.'
        parser.add_argument('-u', '--usetex', action='store_true',
                            default=usetex, dest='usetex', help=hstr)
        hstr = ('White lines on transparent or black background;'
                ' default: black lines on transparent or white background.')
        parser.add_argument('-w', '--white', action='store_true',
                            default=dowhite, dest='dowhite', help=hstr)
        hstr = ('Dots Per inch (DPI) for non-vector output types or rasterized'
                ' maps in vector output (default: 300).')
        parser.add_argument('--dpi', action='store', default=dpi, type=int,
                            dest='dpi', metavar='number', help=hstr)
        hstr = ('Transparent figure background (default: black or white).')
        parser.add_argument('--transparent', action='store_true',
                            default=transparent, dest='transparent', help=hstr)
        parser.add_argument('cargs', nargs='*', default=None,
                            metavar='args', help=iargstr)

        args = parser.parse_args()

        self.args        = args.cargs
        self.plotname    = args.plotname
        self.serif       = args.serif
        self.outtype     = args.outtype
        self.usetex      = args.usetex
        self.dowhite     = args.dowhite
        self.dpi         = args.dpi
        self.transparent = args.transparent

        del parser, args

    # -------------------------------------------------------------------------
    # current layout options
    #
    def set_layout_options(self):
        """
        Standard layout options that can be used in plotting methods

        One can either copy this routine into an extending class and adapt it,
        or add a new method that resets some of the layout options and call it
        in the initialisation of an extending class, or simply set layout
        options in the initialisation of an extending class.

        Current layout options are:

        .. list-table::
           :widths: 15 50
           :header-rows: 1

           * - Option
             - Description
           * - self.nrow
             - number of rows of subplots per figure
           * - self.ncol
             - number of columns of subplots per figure
           * - self.left
             - left space on page
           * - self.right
             - right space on page
           * - self.bottom
             - lower bottom space on page
           * - self.top
             - upper top space on page
           * - self.hspace
             - (horizontal) x-space between subplots
           * - self.vspace
             - (vertical) y-space between subplots
           * - self.textsize
             - standard text size
           * - self.dxabc
             - % of (max-min) shift to the right of left y-axis for a,b,c,...
               labels
           * - self.dyabc
             - % of (max-min) shift up from lower x-axis for a,b,c,... labels
           * - self.lwidth
             - line width
           * - self.elwidth
             - errorbar line width
           * - self.alwidth
             - axis line width
           * - self.msize
             - marker size
           * - self.mwidth
             - marker edge width
           * - self.fgcolor
             - foreground colour
           * - self.bgcolor
             - background colour
           * - self.mcols
             - list of marker colours
           * - self.mcol1
             - marker colour 1
           * - self.mcol2
             - marker colour 2
           * - self.mcol3
             - marker colour 3
           * - self.mcol4
             - marker colour 4
           * - self.mcol5
             - marker colour 5
           * - self.lcol1
             - list of line colours
           * - self.lcol2
             - line colour 1
           * - self.lcol3
             - line colour 2
           * - self.lcol4
             - line colour 3
           * - self.lcol5
             - line colour 4
           * - self.lcols
             - line colour 5
           * - self.ldashes
             - list of line styles
           * - self.llxbbox
             - x-anchor legend bounding box
           * - self.llybbox
             - y-anchor legend bounding box
           * - self.llrspace
             - spacing between rows in legend
           * - self.llcspace
             - spacing between columns in legend
           * - self.llhtextpad
             - pad between the legend handle and text
           * - self.llhlength
             - length of the legend handles
           * - self.frameon
             - if True: draw a frame around the legend.
               If None: use rc
           * - self.dpi
             - DPI of non-vector figure output
           * - self.transparent
             - True for transparent background in figure
           * - self.bbox_inches
             - Bbox in inches. If 'tight', try to figure out the tight bbox of
               the figure
           * - self.pad_inches
             - Amount of padding when bbox_inches is 'tight'

        Examples
        --------
        Setting layout options in initialisation

        .. code-block:: python

           class UsemcPlot(mcPlot):
               def __init__(self, *args, **kwargs):
                   super().__init__(*args, **kwargs)
                   self.lcol1     = 'black'
                   self.mynewcol = 'red'
                   # reset global values after colour changes, etc.
                   self.set_matplotlib_rcparams()

        """
        # layout and spaces
        self.nrow     = 3     # # of rows of subplots per figure
        self.ncol     = 2     # # of columns of subplots per figure
        self.left     = 0.125  # left space on page
        self.right    = 0.9   # right space on page
        self.bottom   = 0.1   # lower space on page
        self.top      = 0.9   # upper space on page
        self.hspace   = 0.1   # x-space between subplots
        self.vspace   = 0.1   # y-space between subplots
        self.textsize = 12    # standard text size
        self.dxabc    = 0.90  # % of (max-min) shift to the right
                              # of left y-axis for a,b,c,... labels
        self.dyabc    = 0.05  # % of (max-min) shift up from lower x-axis
                              # for a,b,c,... labels
        # lines, markers and colours
        self.lwidth  = 1.5  # linewidth
        self.elwidth = 1.0  # errorbar line width
        self.alwidth = 1.0  # axis line width
        self.msize   = 1.5  # marker size
        self.mwidth  = 1.0  # marker edge width
        if self.dowhite:
            self.fgcolor = 'white'
            self.bgcolor = 'black'
        else:
            self.fgcolor = 'black'
            self.bgcolor = 'white'
        # pyjams' amwg color map without the first three colours
        # white, black and purple
        amwg = [(0.141, 0.129, 0.408),  # dark blue
                (0.141, 0.357, 0.6),    # medium blue
                (0.365, 0.533, 0.725),  # light blue
                (0.608, 0.8, 0.835),    # cyan
                (0.0, 0.596, 0.639),    # turquoise
                (0.553, 0.725, 0.259),  # light green
                (0.055, 0.518, 0.345),  # dark green
                (0.937, 0.863, 0.714),  # sand
                (0.808, 0.675, 0.51),   # beige
                (0.957, 0.816, 0.169),  # yellow
                (0.898, 0.6, 0.165),    # orange
                (0.835, 0.294, 0.157),  # light red
                (0.568, 0.176, 0.196)]  # dark red
        self.mcols = amwg
        self.mcol1 = self.mcols[0]   # dark blue
        self.mcol2 = self.mcols[12]  # dark red
        self.mcol3 = self.mcols[2]   # light blue
        self.mcol4 = self.mcols[10]  # orange
        self.mcol5 = self.mcols[6]   # dark green
        self.lcol1 = self.mcol1
        self.lcol2 = self.mcol2
        self.lcol3 = self.mcol3
        self.lcol4 = self.mcol4
        self.lcol5 = self.mcol5
        self.lcols = self.mcols
        # ldashes     = [(5, 2, 2, 2, 2, 2, 2, 2), (2, 2), (10, 3), (5, 3),
        #                (3, 5), (5, 0)]
        self.ldashes = [(5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2),
                        (5, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2),
                        (5, 0),
                        (5, 2),
                        (5, 2, 2, 2),
                        (5, 2, 2, 2, 2, 2),
                        (5, 2, 2, 2, 2, 2, 2, 2),
                        (5, 2, 2, 2, 2, 2, 2, 2, 2, 2)]
        # legend
        self.llxbbox    = 1.0    # x-anchor legend bounding box
        self.llybbox    = 1.0    # y-anchor legend bounding box
        self.llrspace   = 0.     # spacing between rows in legend
        self.llcspace   = 1.0    # spacing between columns in legend
        self.llhtextpad = 0.4    # pad between the legend handle and text
        self.llhlength  = 1.5    # length of the legend handles
        self.frameon    = False  # if True, draw a frame around the legend.
                                 # If None, use rc
        # png
        # self.dpi         = 300
        self.bbox_inches = 'tight'
        self.pad_inches  = 0.035

    # -------------------------------------------------------------------------
    # test figure
    #
    def plot_test(self):
        """
        A simple test plot

        """
        import matplotlib.pyplot as plt

        self.ifig += 1
        iplot  = 0
        print('    Plot - Fig ', self.ifig)
        fig = plt.figure(self.ifig)

        nn = 100
        xx = np.arange(nn) / float(nn) * 4.*np.pi
        yy1 = np.sin(xx)
        yy2 = np.cos(xx)

        xlab = r'4 $\pi$'
        ylab = 'sine and cosine function'
        xlim = None
        ylim = None

        iplot += 1

        sub = fig.add_axes([0.125, 0.667, 0.3375, 0.233], label=str(iplot))

        tarr = ['sin']
        larr = sub.plot(xx, yy1)
        plt.setp(larr[-1], linestyle='-', linewidth=self.lwidth, marker='',
                 color=self.lcols[0])

        tarr += ['cos']
        larr += sub.plot(xx, yy2)
        plt.setp(larr[-1], linestyle='-', linewidth=self.lwidth, marker='',
                 color=self.lcols[-3])

        if xlab != '':
            plt.setp(sub, xlabel=xlab)
        if ylab != '':
            plt.setp(sub, ylabel=ylab)
        sub.grid(False)

        sub.spines['right'].set_color('none')
        sub.spines['top'].set_color('none')

        if xlim is not None:
            plt.setp(sub, xlim=xlim)
        if ylim is not None:
            plt.setp(sub, ylim=ylim)

        ll = sub.legend(larr, tarr, frameon=self.frameon, ncol=1,
                        labelspacing=self.llrspace,
                        handletextpad=self.llhtextpad,
                        handlelength=self.llhlength,
                        loc='upper left',
                        bbox_to_anchor=(self.llxbbox, self.llybbox),
                        scatterpoints=1, numpoints=1)
        plt.setp(ll.get_texts(), fontsize='small')

        # import pdb; pdb.set_trace()

        self.plot_save(fig)

    # -------------------------------------------------------------------------
    # output type
    #
    def set_output_type(self):
        """
        Set type of chosen output.

        Fall back to standard html mode if mlpd3 or hvplot modules are not
        installed.

        """
        self.outtypes = ['', 'pdf', 'png', 'html', 'd3', 'hvplot']
        self.outtype_ends = ['', '.pdf', '_', '.html', '.html', '.html']

        self.outtype  = self.outtype.lower()
        if self.outtype not in self.outtypes:
            estr  = '\nOutput ' + self.outtype + ' type must be in:'
            raise IOError(estr, self.outtypes)

        if (self.outtype == 'd3'):
            try:
                import mpld3
            except ModuleNotFoundError:
                print("    No mpld3 found. Use output type html instead.")
                self.outtype = 'html'

        if (self.outtype == 'hvplot'):
            try:
                import hvplot
            except ModuleNotFoundError:
                raise IOError('Module hvplot not found')
            self.usetex = False

    # -------------------------------------------------------------------------
    # Matplotlib defaults
    #
    def set_matplotlib_rcparams(self):
        """
        Set rcParams depending on output type and other options such as using
        LaTeX, serif fonts, etc.

        rcParams can be overwritten in the initialisation of an extending
        class.

        Current parameters set are:

        .. list-table::
           :widths: 15 50
           :header-rows: 1

           * - Parameter
             - Options
           * - ps
             - papersize, usedistiller
           * - figure
             - figsize, edgecolor, facecolor
           * - text
             - usetex, latex.preamble, color
           * - font
             - family, sans-serif, size
           * - savefig
             - dpi, format, edgecolor, facecolor
           * - axes
             - linewidth, edgecolor, facecolor, labelcolor, prop_cycle
           * - boxplot
             - boxprops.color, capprops.color, flierprops.color,
               flierprops.markeredgecolor, whiskerprops.color
           * - grid
             - color
           * - lines
             - linewidth, color
           * - patch
             - edgecolor
           * - path
             - simplify
           * - xtick
             - color
           * - ytick
             - color

        Examples
        --------
        Setting rcParams in initialisation

        .. code-block:: python

           class UsemcPlot(mcPlot):
               def __init__(self, *args, **kwargs):
                   import matplotlib as mpl
                   super().__init__(*args, **kwargs)
                   mpl.rc('grid', color='red')
                   mpl.rcParams['boxplot.boxprops.color'] = 'blue'

        """
        import matplotlib as mpl
        if (self.outtype == 'pdf'):
            mpl.use('PDF')  # set directly after import matplotlib
            from matplotlib.backends.backend_pdf import PdfPages
            self.PdfPages = PdfPages
            # Customize
            #     http://matplotlib.sourceforge.net/users/customizing.html
            mpl.rc('ps', papersize='a4', usedistiller='xpdf')  # ps2pdf
            mpl.rc('figure', figsize=(8.27, 11.69))  # a4 portrait
            if self.usetex:
                mpl.rc('text', usetex=True)
                if not self.serif:
                    #   r'\usepackage{helvet}',  # use Helvetica
                    mpl.rcParams['text.latex.preamble'] = '\n'.join([
                        # use MyriadPro font
                        r'\usepackage[math,lf,mathtabular,footnotefigures]'
                        r'{MyriadPro}',
                        # normal text font is sans serif
                        r'\renewcommand{\familydefault}{\sfdefault}',
                        r'\figureversion{lining,tabular}',
                        # for permil symbol (load after MyriadPro)
                        r'\usepackage{wasysym}',
                        # for degree symbol (load after MyriadPro)
                        r'\usepackage{gensymb}'])
                else:
                    mpl.rcParams['text.latex.preamble'] = '\n'.join([
                        r'\usepackage{wasysym}',  # for permil symbol
                        r'\usepackage{gensymb}'])  # for degree symbol
            else:
                if self.serif:
                    mpl.rcParams['font.family']     = 'serif'
                    mpl.rcParams['font.sans-serif'] = 'Times'
                else:
                    mpl.rcParams['font.family']     = 'sans-serif'
                    mpl.rcParams['font.sans-serif'] = 'Arial'  # Arial, Verdana
        elif ((self.outtype == 'png') or (self.outtype == 'html') or
              (self.outtype == 'd3') or (self.outtype == 'hvplot')):
            mpl.use('Agg')  # set directly after import matplotlib
            mpl.rc('figure', figsize=(8.27, 11.69))  # a4 portrait
            if self.usetex:
                mpl.rc('text', usetex=True)
                if not self.serif:
                    #   r'\usepackage{helvet}',  # use Helvetica
                    mpl.rcParams['text.latex.preamble'] = '\n'.join([
                        # use MyriadPro font
                        r'\usepackage[math,lf,mathtabular,footnotefigures]'
                        r'{MyriadPro}',
                        # normal text font is sans serif
                        r'\renewcommand{\familydefault}{\sfdefault}',
                        r'\figureversion{lining,tabular}',
                        # for permil symbol (load after MyriadPro)
                        r'\usepackage{wasysym}',
                        # for degree symbol (load after MyriadPro)
                        r'\usepackage{gensymb}'])
                else:
                    mpl.rcParams['text.latex.preamble'] = '\n'.join([
                        r'\usepackage{wasysym}',   # for permil symbol
                        r'\usepackage{gensymb}'])  # for degree symbol
            else:
                if self.serif:
                    mpl.rcParams['font.family']     = 'serif'
                    mpl.rcParams['font.sans-serif'] = 'Times'
                else:
                    mpl.rcParams['font.family']     = 'sans-serif'
                    mpl.rcParams['font.sans-serif'] = 'Arial'  # Arial, Verdana
            mpl.rc('savefig', dpi=self.dpi, format='png')
        else:
            # a4 portrait
            mpl.rc('figure', figsize=(4. / 5. * 8.27, 4. / 5. * 11.69))
        # print(mpl.rcParams)
        mpl.rc('axes', linewidth=self.alwidth, edgecolor=self.fgcolor,
               facecolor=self.bgcolor, labelcolor=self.fgcolor,
               prop_cycle=mpl.rcsetup.cycler('color',
                                             ['8dd3c7', 'feffb3', 'bfbbd9',
                                              'fa8174', '81b1d2', 'fdb462',
                                              'b3de69', 'bc82bd', 'ccebc4',
                                              'ffed6f']))
        mpl.rcParams['boxplot.boxprops.color'] = self.fgcolor
        mpl.rcParams['boxplot.capprops.color'] = self.fgcolor
        mpl.rcParams['boxplot.flierprops.color'] = self.fgcolor
        mpl.rcParams['boxplot.flierprops.markeredgecolor'] = self.fgcolor
        mpl.rcParams['boxplot.whiskerprops.color'] = self.fgcolor
        mpl.rc('figure', edgecolor=self.bgcolor, facecolor=self.bgcolor)
        mpl.rc('font', size=self.textsize)
        mpl.rc('grid', color=self.fgcolor)
        mpl.rc('lines', linewidth=self.lwidth, color=self.fgcolor)
        mpl.rc('patch', edgecolor=self.fgcolor)
        mpl.rc('path', simplify=False)  # do not remove
        mpl.rc('savefig', edgecolor=self.bgcolor, facecolor=self.bgcolor)
        mpl.rc('text', color=self.fgcolor)
        mpl.rc('xtick', color=self.fgcolor)
        mpl.rc('ytick', color=self.fgcolor)

    # -------------------------------------------------------------------------
    # plot begin
    #
    def plot_begin(self):
        import os.path
        """
        Last step of initialisation

        Set output filename depending on chosen output type. Opens output file
        if appropriate.

        """
        if self.plotname == '':
            self.plotfile  = (filebase(os.path.basename(__file__)) +
                              self.outtype_ends[
                                  self.outtypes.index(self.outtype)])
        else:
            self.plotfile = self.plotname

        if self.outtype == '':
            print('    Plot X')
        else:
            print('    Plot', self.plotfile)

        if (self.outtype == 'pdf'):
            self.pdf_pages = self.PdfPages(self.plotfile)
            # figsize = mpl.rcParams['figure.figsize']
        elif self.outtype in ['html', 'd3']:
            print('    Write html file ', self.plotfile)
            self.fhtml = open(self.plotfile, 'w')
            print(f'<!DOCTYPE html PUBLIC "-//W3C//DTD HTML 4.01'
                  f' Transitional//EN">\n'
                  f' "http://www.w3.org/TR/html4/loose.dtd">\n'
                  f'<html>\n'
                  f'    <head>\n'
                  f'        <title>{self.plotfile}</title>\n'
                  f'    </head>\n'
                  f'    <body>\n',
                  file=self.fhtml)
        elif (self.outtype == 'hvplot'):
            print('    Write html file ', self.plotfile)
            with open(self.plotfile, 'w') as ff:
                print(f'<!DOCTYPE html public "-//W3C//DTD HTML'
                      f' 4.01 Frameset//EN"\n'
                      f' "http://www.w3.org/TR/html4/frameset.dtd">\n'
                      f'<html>\n'
                      f'    <head>\n'
                      f'        <title>{self.plotfile}</title>\n'
                      f'    </head>\n'
                      f'    <frameset cols="15%, 85%">\n'
                      f'        <frame name="fixed" src="html/menu.html">\n'
                      f'        <frame name="dynamic" src="html/empty.html">\n'
                      f'    </frameset>\n'
                      f'</html>', file=ff)
            ibase = os.path.dirname(self.plotfile)
            if not ibase:
                ibase = './'
            else:
                ibase += '/'
            if not os.path.exists(ibase + 'html'):
                os.mkdir(ibase + 'html')
            with open(ibase + 'html/empty.html', 'w') as ff:
                print('<!DOCTYPE html public "-//W3C//DTD HTML 4.01'
                      ' Frameset//EN"\n'
                      ' "http://www.w3.org/TR/html4/frameset.dtd">\n'
                      '<html>\n'
                      '    <body>\n'
                      '    </body>\n'
                      '</html>', file=ff)
            self.fhtml = open(ibase + 'html/menu.html', 'w')
            print(f'<!DOCTYPE html public "-//W3C//DTD HTML'
                  f' 4.01 Transitional//EN"\n'
                  f' "http://www.w3.org/TR/html4/loose.dtd">\n'
                  f'<html>\n'
                  f'    <head>\n'
                  f'        <title>{self.plotfile}</title>\n'
                  f'        <base href="html" target="dynamic">\n'
                  f'    </head>\n'
                  f'    <body>\n',
                  file=self.fhtml)

        self.ifig = 0

    def plot_start(self):
        """Alias for plot_begin()"""
        self.plot_begin()

    # -------------------------------------------------------------------------
    # plot save
    #
    def plot_save(self, fig, **kwargs):
        """
        Save figure into output file

        Parameters
        ----------
        fig : matplotlib.figure.Figure
            Matplotlib figure object.    argstr : string, optional
        kwargs : dict, optional
            Additional keywords will be passed to *savefig* for 'pdf',
            'png', and 'html'; to *fig_to_html* for 'd3', and to
            *save* for 'hvplot'. This overwrites *self.transparent*,
            *self.bbox_inches*, *self.pad_inches* for 'pdf', 'png', and 'html'.

        """
        import matplotlib.pyplot as plt
        # save pages
        if (self.outtype == 'pdf'):
            if 'transparent' not in kwargs:
                kwargs.update({'transparent': self.transparent})
            self.pdf_pages.savefig(fig, **kwargs)
            plt.close(fig)
        elif (self.outtype == 'png'):
            pngfile = self.plotfile + "{0:04d}".format(self.ifig) + ".png"
            if 'transparent' not in kwargs:
                kwargs.update({'transparent': self.transparent})
            if 'bbox_inches' not in kwargs:
                kwargs.update({'bbox_inches': self.bbox_inches})
            if 'pad_inches' not in kwargs:
                kwargs.update({'pad_inches': self.pad_inches})
            fig.savefig(pngfile, **kwargs)
            plt.close(fig)
        elif (self.outtype == 'html'):
            pngfile  = filebase(self.plotfile) + "_"
            pngfile += "{0:04d}".format(self.ifig) + ".png"
            if 'transparent' not in kwargs:
                kwargs.update({'transparent': self.transparent})
            if 'bbox_inches' not in kwargs:
                kwargs.update({'bbox_inches': self.bbox_inches})
            if 'pad_inches' not in kwargs:
                kwargs.update({'pad_inches': self.pad_inches})
            fig.savefig(pngfile, **kwargs)
            print('<p><img src=' + pngfile + '></p>', file=self.fhtml)
            plt.close(fig)
        elif (self.outtype == 'd3'):
            import mpld3
            # Does not work:
            #     mpld3.plugins.connect(fig, mpld3.plugins.LinkedBrush(line1))
            d3str = mpld3.fig_to_html(fig, **kwargs)
            print(d3str, file=self.fhtml)
            plt.close(fig)
        elif (self.outtype == 'hvplot'):
            import os
            import inspect
            import hvplot
            import holoviews
            from bokeh.resources import INLINE

            # html directory in same directory as output file
            ibase = os.path.dirname(self.plotfile)
            if not ibase:
                ibase = './'
            else:
                ibase += '/'

            # make holoviews panel or layout
            if 'ncol' in kwargs:
                icol = kwargs['ncol']
            else:
                icol = self.ncol

            # all holoviews charts + layout
            hlist = [ hh[1]
                      for hh in inspect.getmembers(holoviews.element.chart)
                      if inspect.isclass(hh[1]) ]
            hlist.append(holoviews.core.layout.Layout)
            # make holoviews layout if several panels
            if isinstance(fig, (list, tuple)):
                layout = fig[0]
                for pp in fig[1:]:
                    layout = layout + pp
            elif isinstance(layout, hlist):
                layout = fig
            else:
                raise ValueError('figure for hvplot must be list/tuple of'
                                 ' panels, or a holoviews chart or layout.'
                                 ' Given: ' + type(fig))
            if isinstance(layout, holoviews.core.layout.Layout):
                layout = layout.cols(icol)

            # write htmls, menu and plots
            ff = os.path.basename(filebase(self.plotfile))
            # name in side panel
            ffig = f'fig_{self.ifig:04d}'
            # filename in html/
            html = f'{ff}_{ffig}.html'
            # filename in .
            hhtml = ibase + 'html/' + html
            print(f'        <p><a href="{html}" target="dynamic">'
                  f'{ffig}</a>', file=self.fhtml)
            hvplot.save(layout, hhtml, resources=INLINE)

    # -------------------------------------------------------------------------
    # plot end
    #
    def plot_end(self):
        """
        Finish plotting

        Close output file if appropriate or show interactive plots.

        """
        if (self.outtype == 'pdf'):
            self.pdf_pages.close()
        elif (self.outtype == 'png'):
            pass
        elif ( (self.outtype == 'html') or (self.outtype == 'd3') or
               (self.outtype == 'hvplot') ):
            print('    </body>', file=self.fhtml)
            print('</html>', file=self.fhtml)
            self.fhtml.close()
        else:
            import matplotlib.pyplot as plt
            plt.show()

    def plot_stop(self):
        """Alias for plot_end()"""
        self.plot_end()

    def plot_close(self):
        """Alias for plot_end()"""
        self.plot_end()

    def close(self):
        """Alias for plot_end()"""
        self.plot_end()

    def end(self):
        """Alias for plot_end()"""
        self.plot_end()


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
'''
Extension of Python's ConfigParser: adding getlist method.

Using getlist from NoahTheDuke:
    https://gist.github.com/NoahTheDuke/e6d282b421f6a126062e81696e4cfc2a
but extending Python's ConfigParser standard module.


Example
-------
From
    https://www.reddit.com/r/Python/comments/89c9b5/what_config_file_format_do_you_prefer/
File myconfig.ini:
[params]
years: 08,09,10,11,12,13,14,15,16
categories:
    5_000
    12_500
    17_500
    22_500
names: [Steve John Brian]
years and names: ${params:years} ${params:names}

>>> from jconfigparser import jConfigParser
>>> jconfig = jConfigParser()
>>> jconfig.read('myconfig.ini')
>>> jconfig.getiter('params', 'years')
<filter at 0x19ef8641128>

>>> jconfig.getlist('params', 'years')
['08', '09', '10', '11', '12', '13', '14', '15', '16']

>>> list(map(int, jconfig.getiter('params', 'categories')))
[5000, 12500, 17500, 22500]

>>> jconfig.getlist('params', 'names')
['Steve', 'John', 'Brian']

>>> jconfig.getlist('params', 'years and names')
['08', '09', '10', '11', '12', '13', '14', '15', '16', 'Steve', 'John', 'Brian']


License
-------
This file is part of the JAMS Python package, distributed under the MIT License.

Copyright (c) 2018 Matthias Cuntz - mc (at) macu (dot) de

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


History
-------
Written,  Matthias Cuntz, Nov 2018
Modified, Matthias Cuntz, Nov 2018 - getlist, getiter
                                   - sep

'''
try:     # Python 2
    from ConfigParser import ConfigParser
except:  # Python 3
    from configparser import ConfigParser
import itertools as it
import re


__all__ = ['jConfigParser']


# https://stackoverflow.com/questions/1713038/super-fails-with-error-typeerror-argument-1-must-be-type-not-classobj-when
class jConfigParser(ConfigParser, object):
    def __init__(self, *args, **kwargs):
        super(jConfigParser, self).__init__(*args, **kwargs)

    # https://gist.github.com/NoahTheDuke/e6d282b421f6a126062e81696e4cfc2a
    def getiter(self, section, option, sep=r'[ ,\[\]]'):
        r'''
        Return an iterator over the splitted option.
        Separator sep can be given, otherwise pattern '[ ,\[\]]' will be used.

        '''
        lines = self.get(section, option).splitlines()
        if len(lines) == 1:
            return filter(None, re.split(sep, lines[0]))
        else:
            return filter(None, it.chain.from_iterable(
                map(lambda x: re.split(sep, x), lines)))

    def getlist(self, section, option, sep=r'[ ,\[\]]'):
        r'''
        Return a list of the splitted option.
        Separator sep can be given, otherwise pattern '[ ,\[\]]' will be used.

        '''
        lines = self.get(section, option).splitlines()
        if len(lines) == 1:
            return re.split(sep, lines[0])
        else:
            return list(map(lambda x: re.split(sep, x), lines))



#!/usr/bin/env python
import numpy as np


__all__ = ['get_flag', 'get_maxflag', 'set_flag']


# --------------------------------------------------------------------

def get_flag(flags, n):
    """
        Get the flags at position n from CHS data flag vector.


        Definition
        ----------
        def get_flag(flags, n):


        Input
        -----
        flags   ND-array of integers
        n       position to extract


        Output
        ------
        ND-array of integers with flag at position n of flags
        returns -1 if input integer is less than 10**(n+1)


        Examples
        --------
        --> see __init__.py for full example of workflow

        >>> flags = np.array([9, 90, 91, 900, 901, 9001, 9201, 912121212])
        >>> print(get_flag(flags, 0))
        [9 9 9 9 9 9 9 9]

        >>> print(get_flag(flags, 1))
        [-1  0  1  0  0  0  2  1]

        >>> print(get_flag(flags, 2))
        [-1 -1 -1  0  1  0  0  2]

        >>> print(get_flag(flags, 3))
        [-1 -1 -1 -1 -1  1  1  1]

        >>> print(get_flag(flags, 4))
        [-1 -1 -1 -1 -1 -1 -1  2]

        >>> flags = np.array([[9, 90, 91, 900, 901, 9001,  9201, 912121212],
        ...                   [9, 91, 92, 901, 912, 9101, -9999, 912121212]])
        >>> print(get_flag(flags, 1))
        [[-1  0  1  0  0  0  2  1]
         [-1  1  2  0  1  1 -2  1]]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2015
    """

    out = np.ones(flags.shape, dtype=int)*(-1)   # if not enough digits -> -1

    # deal with -9999
    jj = np.where(flags <= 0)
    if jj[0].size>0: out[jj] = -2

    # normal flags starting with 9
    jj = np.where(flags > 0)
    if jj[0].size>0:
        oo = out[jj]
        ilog10 = np.log10(flags[jj]).astype(int) # how many digits
        ii = np.where((ilog10-n) >= 0)              # otherwise extract n-th digit
        if ii[0].size > 0:
            oo[ii] = (flags[jj][ii] // 10**(ilog10[ii]-n)) % 10
        out[jj] = oo

    return out


# --------------------------------------------------------------------

def set_flag(flags, n, iflag, ii=None):
    """
        Set the flags at position n to iflag at indeces ii of CHS data flag vector.


        Definition
        ----------
AL        def set_flag(flags, n, iflag, ii=None):


        Input
        -----
        flags   CHS data flag vector, ND-array of integers
        n       position in flag to set, missing positions will be cretaed and filled with 0
        iflag   set flag to this value


        Optional Input
        --------------
        ii      indices at which to set flag (default: None, i.e. each entry)


        Output
        ------
        ND-array of integers with flag at position n of flags set to iflag


        Examples
        --------
        --> see __init__.py for full example of workflow

        >>> flags = np.array([9, 90, 901, 9101, 912121212, -9999])
        >>> print(set_flag(flags, 1, 2, [0,1,2]))
        [       92        92       921      9101 912121212        90]

        >>> print(set_flag(flags, 3, 2, [0,1,2,3]))
        [     9002      9002      9012      9102 912121212      9000]

        >>> print(set_flag(flags, 1, 2))
        [       92        92       921      9201 922121212        92]

        >>> flags = np.array([[9, 90, 901, 9101, 912121212, -9999],
        ...                   [9, 90, 901, 9101, 912121212, -9999]])
        >>> print(set_flag(flags, 1, 1, [[0,0,1],[0,1,2]]))
        [[       91        91       901      9101 912121212        90]
         [       90        90       911      9101 912121212        90]]

        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Mar 2015
        Modified, MC, Jan 2016 - ND arrays possible
    """
    fflags = flags.copy()
    uu = np.where(fflags < 0)  # if undef, set to treated that mean 9
    if uu[0].size > 0:
        fflags[uu] = 9
    # extend flag vector if needed
    ilog10 = np.log10(fflags).astype(int)  # are there enough flag positions (i)
    jj = np.where(ilog10 < n)
    if jj[0].size > 0:  # increase number of entries (filled with 0)
        fflags[jj] *= 10**(n-ilog10[jj])
        ilog10      = np.log10(fflags).astype(int)

    # get entries in flag vector
    isflags = get_flag(fflags, n)

    # set entries in flag vector
    if ii is None:                                  # set all to iflag
        fflags += 10**(ilog10-n) * (iflag-isflags)
    else:
        if np.size(ii) > 0:                         # set ii to iflag
            fflags[ii] += 10**(ilog10[ii]-n) * (iflag-isflags[ii])

    return fflags


# --------------------------------------------------------------------

def get_maxflag(flags, no=0):
    """
        Returns the overall flag for each flag from CHS data flag vector.


        Definition
        ----------
        def get_maxflag(flags, no):


        Input
        -----
        flags   ND-array of integers
        no      column or list of columns that will be excluded from the determination of the maximum flag


        Output
        ------
        ND-array of integers with overall flag
        returns:
            2, if at least one flag is 2
            1, if at least one flag is 1
            0, if all flags are 0
           -1, if flag is '9'
           -2, if flag is still -9999


        Examples
        --------
        --> see __init__.py for full example of workflow

        >>> flags = np.array([9, 90, 91, 900, 901, 9001, 9201, 912121212, -9999])
        >>> print(get_maxflag(flags))
        [-1  0  1  0  1  1  2  2 -2]

        >>> print(get_maxflag(flags, no=1))
        [-1 -1 -1  0  1  1  1  2 -2]

        >>> print(get_maxflag(flags, no=[1,3]))
        [-1 -1 -1  0  1  0  0  2 -2]

        >>> flags = np.array([[9, 90, 91, 900, 901, 9001,  9201, 912121212],
        ...                   [9, 91, 92, 901, 912, 9101, -9999, 912121212]])
        >>> print(get_maxflag(flags))
        [[-1  0  1  0  1  1  2  2]
         [-1  1  2  1  2  1 -2  2]]

        >>> flags = np.array([[9, 90, 91, 900, 901, 9001,  9201, 912121212],
        ...                   [9, 91, 92, 901, 912, 9101, -9999, 912121212]])
        >>> print(get_maxflag(flags, no=[1,3]))
        [[-1 -1 -1  0  1  0  0  2]
         [-1 -1 -1  1  2  0 -2  2]]


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2015-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  JM, Mar 2015
        Modified, JV, Oct 2015 - no
                  MC, Jan 2016 - no iterable
    """
    # assue no as list
    try:
        nno = len(no)
    except:
        no = [no]

    # Determine overall flag:
    #   maxflag=-2, if flag is '-9999'
    #   maxflag=-1, if flag is '9'
    #   maxflag=0,  if all individual flags are 0, e.g. '9000000'
    #   maxflag=1,  if at least one flag is 1,     e.g. '9000100'
    #   maxflag=2,  if at least one flag is 2,     e.g. '9002010'
    fmax = np.amax(flags)
    if fmax == -9999:
        # return -2 if no flags set yet
        return np.ones(np.shape(flags), dtype=int)*(-2)
    elif fmax == 9:
        # return -1 if no flags set yet but flags were already initialised
        return np.ones(np.shape(flags), dtype=int)*(-1)
    else:
        maxflag = np.ones(np.shape(flags), dtype=int)*(-2)
        n_flags = np.log10(fmax).astype(int)
        for ii in range(n_flags):
            iii = ii+1
            if iii not in no:
                maxflag = np.maximum(maxflag, get_flag(flags,iii))
        return maxflag


# --------------------------------------------------------------------

if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
dec2date : Converts decimal dates to calendar dates.

This module was written by Arndt Piayda and then enhanced and
maintained by Matthias Cuntz while at Department of Computational
Hydrosystems, Helmholtz Centre for Environmental Research - UFZ,
Leipzig, Germany, and continued by Matthias Cuntz while at Institut
National de Recherche pour l'Agriculture, l'Alimentation et
l'Environnement (INRAE), Nancy, France.

Copyright (c) 2010-2020 Matthias Cuntz - mc (at) macu (dot) de
Released under the MIT License; see LICENSE file for details.

* Written Jun 2010 by Arndt Piayda
* Input can be scalar, list, array, or mix of it, Feb 2012, Matthias Cuntz
* Changed checks, added calendars decimal and decimal360, Feb 2012, Matthias Cuntz
* fulldate=True default, Feb 2012, Matthias Cuntz
* Change keyword name: units -> refdate, Jun 2012, Matthias Cuntz
* units has now same meaning as in netcdftime, Jun 2012, Matthias Cuntz
* Include cdo absolute date units 'day as %Y%m%d.%f', Jun 2012, Matthias Cuntz
* Solved Excel leap year problem, Feb 2013, Matthias Cuntz
* Ported to Python 3, Feb 2013, Matthias Cuntz
* Small bug in eng output, May 2013, Arndt Piayda
* Excel starts at 1 not at 0 on 01 January 1900 or 1904, Oct 2013, Matthias Cuntz
* Bug: 01.01.0001 was substracted if Julian calendar even with units given, Oct 2013, Matthias Cuntz
* Include units 'month as %Y%m.%f' and 'year as %Y.%f', May 2016, Matthias Cuntz
* Adapted to new netCDF4/netcdftime (>= v1.0) and datetime (>= Python v2.7.9), Jun 2015, Matthias Cuntz
* leap always integer, Oct 2016, Matthias Cuntz
* Using numpy docstring format, May 2020, Matthias Cuntz
* Succeed eng by en keyword as in ascii2ascii and date2dec, Jul 2020, Matthias Cuntz
* proleptic_gregorian instead of gregorian calendar for Excel dates, Jul 2020, Matthias Cuntz

.. moduleauthor:: Matthias Cuntz, Arndt Piayda

The following functions are provided

.. autosummary::
   dec2date
"""
from __future__ import division, absolute_import, print_function
from __future__ import division, absolute_import, print_function
from warnings import warn, filterwarnings
filterwarnings("default", category=DeprecationWarning)
import numpy as np


__all__ = ['dec2date']


def dec2date(indata, calendar='standard', refdate=None, units=None,
             excelerr=True, fulldate=None, yr=False,
             mo=False, dy=False, hr=False, mi=False,
             sc=False, ascii=False, en=False, eng=False):
    """
    Converts scale and array_like with decimal dates into
    calendar dates. Supported time formats are:
    standard, gregorian, julian, proleptic_gregorian,
    excel1900, excel1904, 365_day, noleap, 366_day, all_leap,
    and 360_day.

    Input is decimal date in units of days.

    Output is year, month, day, hour, minute, second
    or any combination of them. Output in string format is possible.


    Parameters
    ----------
    indata : array_like
        Input decimal dates. Dates must be positive.
    calendar : str, optional
        Calendar of input dates (default: 'standard').

        Possible values are:

        'standard', 'gregorian' = julian calendar from
        01.01.-4712 12:00:00 (BC) until 05.03.1583 00:00:00 and
        gregorian calendar from 15.03.1583 00:00:00 until now.
        Missing 10 days do not exsist.

        'julian' = julian calendar from 01.01.-4712 12:00:00 (BC)
         until now.

        'proleptic_gregorian' = gregorian calendar from
        01.01.0001 00:00:00 until now.

        'excel1900' = Excel dates with origin at
        01.01.1900 00:00:00.

        'excel1904' = Excel 1904 (Lotus) format.
        Same as excel1904 but with origin at
        01.01.1904 00:00:00.

        '365_day', 'noleap' = 365 days format,
        i.e. common years only (no leap years)
        with origin at 01.01.0001 00:00:00.

        '366_day', 'all_leap' = 366 days format,
        i.e. leap years only (no common years)
        with origin at 01.01.0001 00:00:00.

        '360_day' = 360 days format,
        i.e. years with only 360 days (30 days per month)
        with origin at 01.01.0001 00:00:00.

        'decimal' = decimal year instead of decimal days.

        'decimal360' = decimal year with a year of 360 days, i.e. 12 month with 30 days each.


    Optional Arguments
    ------------------
    refdate : str, optional
        Reference date for 'days since refdate' can be set by user. Input must be a
        string in the format 'yyyy-mm-dd hh:mm:ss'.
        Default values for different `calendars` are set automatically.
    units : str, optional
        Units of the the time stamp can be given. This can only be the following two:

        'day as %Y%m%d.%f', or

        'days since refdate', with refdate in the format 'yyyy-mm-dd hh:mm:ss'.

        If `units='day as %Y%m%d.%f'` then `calendar` is ignored and the dates will be returned
        assuming that %f is fractional date from 00:00:00 h.
    excelerr : bool, optional
       In Excel, the year 1900 is normally considered a leap year,
       which it was not. By default, this error is taken into account
       if calendar='excel1900' (default: True).

       1900 is not considered a leap year if excelerr=False.


    Returns
    -------
    list of array_like
        `fulldate` -> output arrays with year, month, day, hour, minute, second

                      Default fulldate is overwritten by selection of special output `yr`,`mn`,`dy`,`hr`,`mi`,`sc`

        `yr`       -> output array with year

        `mo`       -> output array with month

        `dy`       -> output array with day

        `hr`       -> output array with hour

        `mi`       -> output array with minute

        `sc`       -> output array with second

        `ascii`    -> output array with strings of the format 'dd.mm.yyyy hh:mm:ss'

        `en`      -> output array with strings of the format 'yyyy-mm-dd hh:mm:ss'

        `eng`     -> Same as en: obsolete.

    Notes
    -----
    Most versions of `datetime do` not support negative years,
    i.e. Julian days < 1721423.5 = 01.01.0001 00:00.

    There is an issue in `netcdftime` version < 0.9.5 in proleptic_gregorian for dates before year 301:
      dec2date(date2dec(ascii='01.01.0300 00:00:00', calendar='proleptic_gregorian'), calendar='proleptic_gregorian')
        [300, 1, 2, 0, 0, 0]
      dec2date(date2dec(ascii='01.01.0301 00:00:00', calendar='proleptic_gregorian'), calendar='proleptic_gregorian')
        [301, 1, 1, 0, 0, 0]

    Requires `netcdftime.py` from module netcdftime available at:
        http://netcdf4-python.googlecode.com

    Examples
    --------
    # Some implementations of datetime have problems with negative years
    >>> import datetime
    >>> if datetime.MINYEAR > 0:
    ...     print('The minimum year in your datetime implementation is ', datetime.MINYEAR)
    ...     print('i.e. it does not support negative years (BC).')
    The minimum year in your datetime implementation is  1
    i.e. it does not support negative years (BC).

    #calendar = 'standard'
    >>> year   = np.array([2000,1810,1630,1510,1271,619,1])
    >>> month  = np.array([1,4,7,9,3,8,1])
    >>> day    = np.array([5,24,15,20,18,27,1])
    >>> hour   = np.array([12,16,10,14,19,11,12])
    >>> minute = np.array([30,15,20,35,41,8,0])
    >>> second = np.array([15,10,40,50,34,37,0])
    >>> from date2dec import date2dec
    >>> decimal = date2dec(calendar='standard', yr=year, mo=month, dy=day, hr=hour, mi=minute, sc=second)
    >>> year1, month1, day1, hour1, minute1, second1 = dec2date(decimal, calendar= 'standard', fulldate=True)
    >>> print(year1)
    [2000 1810 1630 1510 1271  619    1]
    >>> print(month1)
    [1 4 7 9 3 8 1]
    >>> print(day1)
    [ 5 24 15 20 18 27  1]
    >>> print(hour1)
    [12 16 10 14 19 11 12]
    >>> print(minute1)
    [30 15 20 35 41  8  0]
    >>> print(second1)
    [15 10 40 50 34 37  0]

    # calendar = 'julian'
    >>> decimal = date2dec(calendar='julian', yr=year, mo=month, dy=day, hr=hour, mi=minute, sc=second)
    >>> year1 = dec2date(decimal, calendar='julian', yr=True)
    >>> print(year1)
    [2000 1810 1630 1510 1271  619    1]

    # calendar = 'proleptic_gregorian'
    >>> decimal = date2dec(calendar='proleptic_gregorian', yr=year, mo=month, dy=day, hr=hour, mi=minute, sc=second)
    >>> ascii = dec2date(decimal, calendar='proleptic_gregorian', ascii=True)
    >>> print(ascii[::4])
    ['05.01.2000 12:30:15' '18.03.1271 19:41:34']

    # calendar = 'excel1900' WITH excelerr = True -> 1900 considered as leap year
    >>> decimal = date2dec(calendar='excel1900', yr=year, mo=month, dy=day, hr=hour, mi=minute, sc=second)
    >>> year1, day1 = dec2date(decimal, calendar='excel1900', yr=True, dy=True)
    >>> print(year1)
    [2000 1810 1630 1510 1271  619    1]
    >>> print(day1)
    [ 5 24 15 20 18 27  1]

    # calendar = 'excel1900' WITH excelerr = False -> 1900 considered as NO leap year
    # Older versions of netcdftime.py produced unnecessary output (Line 262)
    # >>> decimal = date2dec(calendar='excel1900',yr=year,mo=month,dy=day,hr=hour,mi=minute,sc=second,excelerr=False)
    # >>> if nt.__version__ < '0.9.4':
    # ...     asciidate = dec2date(decimal, calendar='excel1900', ascii = True, excelerr = False)
    # ... elif nt.__version__ == '0.9.4':
    # ...     asciidate = dec2date(decimal, calendar='excel1900', ascii = True, excelerr = False)
    # ...     for i in range(3):
    # ...         print('0 300')
    # ... else:
    # ...     asciidate = dec2date(decimal, calendar='excel1900', ascii = True, excelerr = False)
    # ...     for i in range(7):
    # ...         print('0 300')
    # 0 300
    # 0 300
    # 0 300
    # 0 300
    # 0 300
    # 0 300
    # 0 300

    # >>> print(asciidate[::4])
    # ['05.01.2000 12:30:15' '18.03.1271 19:41:34']

    # calendar = 'excel1904'
    >>> decimal = date2dec(calendar='excel1904', yr=year, mo=month, dy=day, hr=hour, mi=minute, sc=second)
    >>> asciidate = dec2date(decimal, calendar='excel1904', ascii = True)
    >>> print(asciidate[::4])
    ['05.01.2000 12:30:15' '18.03.1271 19:41:34']
    >>> asciidate = dec2date(decimal, calendar='excel1904', ascii=True, refdate='1909-12-31 00:00:00')
    >>> print(asciidate[::4])
    ['05.01.2006 12:30:15' '18.03.1277 19:41:34']
    >>> print(dec2date(decimal[::4], calendar='excel1904', ascii=True, units='days since 1909-12-31 00:00:00'))
    ['05.01.2006 12:30:15' '18.03.1277 19:41:34']
    >>> print(dec2date(decimal[::4], calendar='excel1904', ascii=True, units='days since 1910-01-01 00:00:00'))
    ['06.01.2006 12:30:15' '19.03.1277 19:41:34']

    # check especially 1900 (no) leap year in Excel
    >>> year1   = np.array([1900,1900,1900,1900])
    >>> month1  = np.array([2,2,3,1])
    >>> day1    = np.array([28,29,1,1])
    >>> decimal = date2dec(calendar='excel1900', yr=year1, mo=month1, dy=day1)
    >>> month2, day2 = dec2date(decimal, calendar='excel1900', mo=True, dy=True)
    >>> print(month2)
    [2 2 3 1]
    >>> print(day2)
    [28 29  1  1]
    >>> decimal = date2dec(calendar='excel1900', yr=year1, mo=month1, dy=day1, excelerr=False)
    >>> month2, day2 = dec2date(decimal, calendar='excel1900', mo=True, dy=True, excelerr=False)
    >>> print(month2)
    [2 3 3 1]
    >>> print(day2)
    [28  1  1  1]
    >>> decimal = date2dec(calendar='excel1904', yr=year1, mo=month1, dy=day1)
    >>> month2, day2 = dec2date(decimal, calendar='excel1904', mo=True, dy=True)
    >>> print(month2)
    [2 3 3 1]
    >>> print(day2)
    [28  1  1  1]

    # calendar = '365_day'
    >>> decimal = date2dec(calendar='365_day',yr=year,mo=month,dy=day,hr=hour,mi=minute,sc=second)
    >>> asciidate = dec2date(decimal, calendar='365_day', ascii = True)
    >>> print(asciidate[::4])
    ['05.01.2000 12:30:15' '18.03.1271 19:41:34']

    # calendar = '366_day'
    >>> decimal = date2dec(calendar='366_day',yr=year,mo=month,dy=day,hr=hour,mi=minute,sc=second)
    >>> asciidate = dec2date(decimal, calendar='366_day', ascii = True)
    >>> print(asciidate[::4])
    ['05.01.2000 12:30:15' '18.03.1271 19:41:34']

    # calendar = '360_day'
    >>> decimal = date2dec(calendar='360_day',yr=year,mo=month,dy=day,hr=hour,mi=minute,sc=second)
    >>> asciidate = dec2date(decimal, calendar='360_day', ascii = True)
    >>> print(asciidate[::4])
    ['05.01.2000 12:30:15' '18.03.1271 19:41:34']

    >>> print(dec2date(719644.52101, calendar='proleptic_gregorian', ascii = True))
    28.04.1971 12:30:15
    >>> dec = date2dec(ascii='02.03.1910 03:44:55', calendar='decimal')
    >>> print(dec2date(dec, calendar='decimal', ascii=True))
    02.03.1910 03:44:55
    >>> dec = date2dec(ascii='02.03.1910 03:44:55', calendar='decimal360')
    >>> print(dec2date(dec, calendar='decimal360', ascii=True))
    02.03.1910 03:44:55
    >>> print(dec2date([dec,dec], calendar='decimal360', ascii=True))
    ['02.03.1910 03:44:55', '02.03.1910 03:44:55']
    >>> print(dec2date([[dec,dec],[dec,dec],[dec,dec]], calendar='decimal360', ascii=True)[0])
    ['02.03.1910 03:44:55', '02.03.1910 03:44:55']
    >>> print(dec2date(np.array([dec,dec]), calendar='decimal360', ascii=True))
    ['02.03.1910 03:44:55' '02.03.1910 03:44:55']
    >>> print(dec2date(np.array([[dec,dec],[dec,dec],[dec,dec]]), calendar='decimal360', ascii=True)[0:2,0])
    ['02.03.1910 03:44:55' '02.03.1910 03:44:55']

    >>> absolut = np.array([20070102.0034722, 20070102.0069444])
    >>> print(dec2date(absolut, units='day as %Y%m%d.%f', ascii=True))
    ['02.01.2007 00:05:00' '02.01.2007 00:10:00']
    >>> absolut = [20070102.0034722, 20070102.0069444]
    >>> print(dec2date(absolut, units='day as %Y%m%d.%f', ascii=True))
    ['02.01.2007 00:05:00', '02.01.2007 00:10:00']

    >>> absolut = np.array([200401.5, 200402.5, 201011.5, 201002.5])
    >>> print(dec2date(absolut, units='month as %Y%m.%f', ascii=True))
    ['15.01.2004 12:00:00' '14.02.2004 12:00:00' '15.11.2010 00:00:00' '14.02.2010 00:00:00']

    >>> absolut = np.array([2004.5, 2010.5])
    >>> print(dec2date(absolut, units='year as %Y.%f', ascii=True))
    ['01.07.2004 00:00:00' '01.07.2010 12:00:00']

    # en, eng
    >>> print(dec2date(719644.52101, calendar='proleptic_gregorian', en=True))
    1971-04-28 12:30:15
    >>> print(dec2date(719644.52101, calendar='proleptic_gregorian', eng=True))
    1971-04-28 12:30:15

    History
    -------
    Written  Arndt Piayda,   Jun 2010
    Modified Matthias Cuntz, Feb 2012 - Input can be scalar or array
                                      - Default: fulldate=True
                                      - Changed checks for easier extension
                                      - decimal, decimal360
             Matthias Cuntz, Jun 2012 - former units keyword is now called refdate
                                      - units has now original meaning as in netcdftime
                                      - units='day as %Y%m%d.%f'
             Matthias Cuntz, Feb 2013 - solved Excel leap year problem.
             Matthias Cuntz, Feb 2013 - ported to Python 3
             Arndt Piayda,   May 2013 - solved eng output problem.
             Matthias Cuntz, Oct 2013 - Excel starts at 1 not at 0
             Matthias Cuntz, Oct 2013 - units bugs, e.g. 01.01.0001 was substracted if Julian calendar even with units
             Matthias Cuntz, May 2016 - units=='month as %Y%m.%f', units=='year as %Y.%f'
             Matthias Cuntz, Oct 2016 - netcdftime provided even with netCDF4 > 1.0.0; make leap always integer
             Matthias Cuntz, May 2020 - numpy docstring format
             Matthias Cuntz, Jul 2020 - en for eng
             Matthias Cuntz, Jul 2020 - use proleptic_gregorian for Excel dates
             Matthias Cuntz, May 2021 - rm np.int and np.float
                                      - (almost) flake8 compatible
    """
    warn('The function dec2date is deprecated and superseded by'
         ' num2date/dec2date.', category=DeprecationWarning)
    #
    # Constants
    calendars = ['standard', 'gregorian', 'julian', 'proleptic_gregorian',
                 'excel1900', 'excel1904', '365_day', 'noleap', '366_day',
                 'all_leap', '360_day', 'decimal', 'decimal360']
    #
    # Checks
    import netCDF4 as nt
    try:
        tst = nt.date2num
        tst = nt.datetime
    except:
        try:
            import netcdftime as nt
            if ((nt.__version__ <= '0.9.2') & (calendar == '360_day')):
                raise ValueError(
                    "date2dec error: Your version of netcdftime.py is equal"
                    " or below 0.9.2. The 360_day calendar does not work with"
                    " arrays here. Please download a newer one.")
        except:
            import cftime as nt
    #
    calendar = calendar.lower()
    if (calendar not in calendars):
        raise ValueError("dec2date error: Wrong calendar! Choose: " +
                         ''.join([i + ' ' for i in calendars]))
    if refdate and units:
        raise ValueError("dec2date error: either refdate or units can be given.")
    # obsolete eng
    if en and eng:
        raise ValueError("dec2date error: 'eng' was succeeded by 'en'. Only one can be given.")
    if eng and (not en):
        en = eng
    #
    # Default
    if np.sum(np.array([yr, mo, dy, hr, mi, sc])) >= 1:
        ii = True
    else:
        ii = False
    if ((ascii | en | ii) and (not fulldate)):
        fulldate = False
    if ((not (ascii | en | ii)) and (not fulldate)):
        fulldate = True
    if fulldate:
        yr = True
        mo = True
        dy = True
        hr = True
        mi = True
        sc = True
        # Further checks
    if np.sum(np.array([ascii, fulldate, en])) > 1:
        raise ValueError("dec2date error: Only one of ascii, fulldate, or en can be chosen.")
    if np.sum(np.array([ascii, en, ii])) > 1:
        raise ValueError("dec2date error: If ascii, fulldate or en then no special selection yr,mo,dy,hr,mi,sc possible.")
    #
    # Input size and shape
    islist = type(indata) != type(np.array(indata))
    isarr = np.ndim(indata)
    if (islist & (isarr > 2)):
        raise ValueError("dec2date error: input is list > 2D; Use array input")
    if isarr == 0:
        indata = np.array([indata])
    else:
        indata = np.array(indata)
    insize  = indata.size
    inshape = indata.shape
    indata  = indata.flatten()
    #
    # depending on chosen calendar and optional set of the time refdate
    # calendar date is calculated
    if units == 'day as %Y%m%d.%f':
        fdy    = indata % 1.  # day fraction
        indata = indata - fdy
        day    = np.rint(indata % 100.).astype(int)
        indata = indata - day
        tmp    = indata % 10000.
        month  = np.rint(tmp / 100.).astype(int)
        indata = indata - tmp
        year   = np.rint(indata / 10000.).astype(int)
        secs   = np.rint(fdy * 86400.)
        hour   = np.floor(secs / 3600.).astype(int)
        secs   = secs - 3600. * hour
        minute = np.floor(secs / 60.).astype(int)
        second = np.rint(secs - 60. * minute).astype(int)
    elif units == 'month as %Y%m.%f':
        fmo    = indata % 1.  # month fraction
        indata = indata - fmo
        month  = np.rint(indata % 100.).astype(int)
        indata = indata - month
        year   = np.rint(indata / 100.).astype(int)
        leap   = np.where((((year % 4) == 0) & ((year % 100) != 0)) |
                          ((year % 400) == 0), 1, 0)
        dim    = np.array([[-9, 31, 28, 31, 30, 31, 30,
                            31, 31, 30, 31, 30, 31],
                           [-9, 31, 29, 31, 30, 31, 30,
                            31, 31, 30, 31, 30, 31]])
        indata = dim[(leap, month)] * fmo
        fdy    = indata % 1.  # day fraction
        indata = indata - fdy
        day    = np.rint(indata % 100.).astype(int)
        secs   = np.rint(fdy * 86400.)
        hour   = np.floor(secs / 3600.).astype(int)
        secs   = secs - 3600. * hour
        minute = np.floor(secs / 60.).astype(int)
        second = np.rint(secs - 60. * minute).astype(int)
    elif units == 'year as %Y.%f':
        fyr    = indata % 1.  # year fraction
        year   = np.rint(indata - fyr).astype(int)
        leap   = np.where((((year % 4) == 0) & ((year % 100) != 0)) |
                          ((year % 400) == 0), 1, 0)
        dsiy   = np.array([365, 366])
        indata = dsiy[leap] * fyr
        fdy    = indata % 1.  # day fraction
        doy    = np.rint(indata - fdy).astype(int)
        diy    = np.array([ [-9, 0, 31, 59, 90, 120, 151, 181,
                             212, 243, 273, 304, 334, 365],
                            [-9, 0, 31, 60, 91, 121, 152, 182,
                             213, 244, 274, 305, 335, 366] ])
        month = np.zeros(insize, dtype=int)
        day   = np.zeros(insize, dtype=int)
        for i in range(insize):
            month[i] = np.where(doy[i] > np.squeeze(diy[leap[i], :]))[0][-1]
            day[i]   = doy[i] - diy[leap[i], month[i]]
        secs   = np.rint(fdy * 86400.)
        hour   = np.floor(secs / 3600.).astype(int)
        secs   = secs - 3600. * hour
        minute = np.floor(secs / 60.).astype(int)
        second = np.rint(secs - 60. * minute).astype(int)
    else:
        if (calendar == 'standard') or (calendar == 'gregorian'):
            dec0 = 0
            if units:
                unit = units
            elif refdate:
                # unit = 'days since %s' % (refdate)
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 0001-01-01 12:00:00'
                dec0 = 1721424
            timeobj = nt.num2date(indata - dec0, unit, calendar='gregorian')
        elif calendar == 'julian':
            dec0 = 0
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 0001-01-01 12:00:00'
                dec0 = 1721424
            timeobj = nt.num2date(indata - dec0, unit, calendar='julian')
        elif calendar == 'proleptic_gregorian':
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 0001-01-01 00:00:00'
            timeobj = nt.num2date(indata, unit, calendar='proleptic_gregorian')
        elif calendar == 'excel1900':
            doerr = False
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 1899-12-31 00:00:00'
                if excelerr:
                    doerr = True
            if doerr:
                indata1 = np.where(indata >= 61., indata - 1, indata)
                timeobj = nt.num2date(indata1, unit,
                                      calendar='proleptic_gregorian')
            else:
                timeobj = nt.num2date(indata, unit,
                                      calendar='proleptic_gregorian')
        elif calendar == 'excel1904':
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 1903-12-31 00:00:00'
            timeobj = nt.num2date(indata, unit, calendar='proleptic_gregorian')
        elif (calendar == '365_day') or (calendar == 'noleap'):
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 0001-01-01 00:00:00'
            timeobj = nt.num2date(indata, unit, calendar='365_day')
        elif (calendar == '366_day') or (calendar == 'all_leap'):
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 0001-01-01 00:00:00'
            timeobj = nt.num2date(indata, unit, calendar='366_day')
        elif calendar == '360_day':
            if units:
                unit = units
            elif refdate:
                unit = 'days since {0:s}'.format(refdate)
            else:
                unit = 'days since 0001-01-01 00:00:00'
            timeobj = nt.num2date(indata, unit, calendar='360_day')
        elif calendar == 'decimal':
            fyear = np.trunc(indata)
            year  = np.array(fyear, dtype=int)
            leap  = ((((year % 4) == 0) & ((year % 100) != 0)) |
                     ((year % 400) == 0)).astype(int)
            fleap = leap.astype(float)
            fract_date = indata - fyear
            days_year  = 365.
            # date in hours
            fhoy  = fract_date * (days_year + fleap) * 24.
            fihoy = np.trunc(fhoy)
            ihoy  = np.array(fihoy, dtype=int)
            # minutes
            fmoy    = (fhoy - fihoy) * 60.
            fminute = np.trunc(fmoy)
            minute  = np.array(fminute, dtype=int)
            # seconds
            second = np.array(np.trunc((fmoy - fminute) * 60.), dtype=int)
            # months
            fdoy  = (fihoy / 24.) + 1.
            fidoy = np.trunc(fdoy)
            idoy  = np.array(fidoy, dtype=int)
            month = np.zeros(insize, dtype=int)
            diy   = np.array([ [-9, 0, 31, 59, 90, 120, 151, 181,
                                212, 243, 273, 304, 334, 365],
                               [-9, 0, 31, 60, 91, 121, 152, 182,
                                213, 244, 274, 305, 335, 366] ])
            for i in range(insize):
                ii = np.squeeze(
                    np.where(idoy[i] > np.squeeze(diy[leap[i], :])))
                month[i] = ii[-1]
            # days
            fday = np.zeros(insize, dtype=float)
            for i in range(insize):
                fday[i] = np.trunc(fdoy[i] - float(diy[leap[i], month[i]]))
            day = np.array(fday, dtype=int)
            # hours
            hour = ihoy % 24
        elif calendar == 'decimal360':
            fyear = np.trunc(indata)
            year  = np.array(fyear, dtype=int)
            fract_date = indata - fyear
            days_year  = 360.
            # date in hours
            fhoy  = fract_date * days_year * 24.
            fihoy = np.trunc(fhoy)
            ihoy  = np.array(fihoy, dtype=int)
            # minutes
            fmoy    = (fhoy - fihoy) * 60.
            fminute = np.trunc(fmoy)
            minute  = np.array(fminute, dtype=int)
            # seconds
            second = np.array(np.trunc((fmoy - fminute) * 60.), dtype=int)
            # months
            fdoy  = (fihoy / 24.) + 1.
            fidoy = np.trunc(fdoy)
            idoy  = np.array(fidoy, dtype=int)
            month = np.zeros(insize, dtype=int)
            diy   = np.array([ -9, 0, 30, 60, 90, 120, 150, 180,
                               210, 240, 270, 300, 330, 360 ])
            for i in range(insize):
                ii = np.squeeze(np.where(idoy[i] > diy))
                month[i] = ii[-1]
            # days
            fday = np.zeros(insize, dtype=float)
            for i in range(insize):
                fday[i] = np.trunc(fdoy[i] - float(diy[month[i]]))
            day = np.array(fday, dtype=int)
            # hours
            hour = ihoy % 24
        else:
            raise ValueError("dec2date error: calendar not implemented; should have been catched before.")

        if (calendar not in ['decimal', 'decimal360']):
            timeobjfl = timeobj.flatten()
            year   = np.array([timeobjfl[i].year
                               for i in range(insize)], dtype=int)
            month  = np.array([timeobjfl[i].month
                               for i in range(insize)], dtype=int)
            day    = np.array([timeobjfl[i].day
                               for i in range(insize)], dtype=int)
            hour   = np.array([timeobjfl[i].hour
                               for i in range(insize)], dtype=int)
            minute = np.array([timeobjfl[i].minute
                               for i in range(insize)], dtype=int)
            second = np.array([timeobjfl[i].second
                               for i in range(insize)], dtype=int)
            if (calendar == 'excel1900') & excelerr:
                ii  = np.where((indata >= 60.) & (indata < 61.))[0]
                if np.size(ii) > 0:
                    month[ii] = 2
                    day[ii]   = 29
    #
    # Ascii output
    if ascii:
        output = [ '%02d.%02d.%04d %02d:%02d:%02d' %
                   (day[i], month[i], year[i], hour[i], minute[i], second[i])
                   for i in range(insize) ]
        output = np.reshape(output, inshape)
        if isarr == 0:
            output = output[0]
    # Ascii english output
    elif en:
        output = ['%04d-%02d-%02d %02d:%02d:%02d' %
                  (year[i], month[i], day[i], hour[i], minute[i], second[i])
                  for i in range(insize) ]
        output = np.reshape(output, inshape)
        if isarr == 0:
            output = output[0]
    else:
        # Individual output
        # if one, some or all of yr, mo, dy, hr, mi or sc is
        # choosen by the user as output, arrays for datetime
        year   = np.reshape(year, inshape)
        month  = np.reshape(month, inshape)
        day    = np.reshape(day, inshape)
        hour   = np.reshape(hour, inshape)
        minute = np.reshape(minute, inshape)
        second = np.reshape(second, inshape)
        if isarr == 0:
            year   = int(year)
            month  = int(month)
            day    = int(day)
            hour   = int(hour)
            minute = int(minute)
            second = int(second)
        # filling of output list:
        output = []
        if yr:
            output += [year]
        if mo:
            output += [month]
        if dy:
            output += [day]
        if hr:
            output += [hour]
        if mi:
            output += [minute]
        if sc:
            output += [second]
        # return output arrays:
        if len(output) == 1:
            output = output[0]

    if isarr != 0:
        if islist:
            ns = np.size(inshape)
            if ns == 1:
                output = [i for i in output]
            elif ns == 2:
                loutput = [ i for i in output[:, 0]]
                for i in range(np.size(output[:, 0])):
                    loutput[i] = list(np.squeeze(output[i, :]))
                output = loutput
            else:
                raise ValueError("dec2date error: list output > 2D; should have been catched before.")

    return output


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
import os
import datetime
import time
from pyjams.argsort import argsort


__all__ = ['get_binary', 'get_check_binary',
           'get_mac_ascii', 'get_check_mac_ascii',
           'get_unix_ascii', 'get_check_unix_ascii',
           'get_windows_ascii', 'get_check_windows_ascii',
           'get_names', 'get_names_dates', 'get_names_sizes',
           'get_names_times',
           'get_names_dates_sizes', 'get_names_times_sizes',
           'get_size', 'get_sizes',
           'set_mtime',
           'put_binary', 'put_check_binary']


# ------------------------------------------------------------------------------------------

def get_binary(ftp, fname):
    """
        Get a binary file from an open FTP connection.


        Definition
        ----------
        def get_binary(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        get_binary(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                                 http://www.java2s.com/Tutorial/Python/0420__Network/Binaryfiledownload.htm
    """
    ftp.voidcmd("TYPE I")
    datasock, estsize = ftp.ntransfercmd("RETR "+fname)
    transbytes = 0
    fd = open(fname, 'wb')
    while True:
        buf = datasock.recv(2048)
        if not len(buf):
            break
        fd.write(buf)
        transbytes += len(buf)
    fd.close()
    datasock.close()
    ftp.voidresp()
    set_mtime(ftp, fname)

# ------------------------------------------------------------------------------------------

def get_check_binary(ftp, fname):
    """
        Get a binary file from an open FTP connection and
        check that local filesize is the same as the remote file size.


        Definition
        ----------
        def get_check_binary(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        if not get_check_binary(ftp, 'test.dat'):
            print('Error in transfer')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    rsize = get_size(ftp, fname)
    get_binary(ftp, fname)
    lsize = os.stat(fname).st_size
    if rsize == lsize:
        return True
    else:
        return False

# ------------------------------------------------------------------------------------------

def get_mac_ascii(ftp, fname):
    """
        Get a Mac ascii file from an open FTP connection.


        Definition
        ----------
        def get_mac_ascii(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        get_mac_ascii(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                                 http://www.java2s.com/Tutorial/Python/0420__Network/ASCIIfiledownload.htm
    """
    def writeline(data):
        fd.write(data + '\r') # Mac-Format
    fd = open(fname, 'wt')
    ftp.retrlines('RETR '+fname, writeline)
    fd.close()
    set_mtime(ftp, fname)

# ------------------------------------------------------------------------------------------

def get_check_mac_ascii(ftp, fname):
    """
        Get a Mac ascii file from an open FTP connection and
        check that local filesize is the same as the remote file size.


        Definition
        ----------
        def get_check_mac_ascii(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        if not get_check_mac_ascii(ftp, 'test.dat'):
            print('Error in transfer')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    rsize = get_size(ftp, fname)
    get_mac_ascii(ftp, fname)
    lsize = os.stat(fname).st_size
    if rsize == lsize:
        return True
    else:
        return False

# ------------------------------------------------------------------------------------------

def get_unix_ascii(ftp, fname):
    """
        Get a Unix/Linux ascii file from an open FTP connection.


        Definition
        ----------
        def get_unix_ascii(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        get_unix_ascii(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                                 http://www.java2s.com/Tutorial/Python/0420__Network/ASCIIfiledownload.htm
    """
    def writeline(data):
        fd.write(data + '\n') # Unix-Format
    fd = open(fname, 'wt')
    ftp.retrlines('RETR '+fname, writeline)
    fd.close()
    set_mtime(ftp, fname)

# ------------------------------------------------------------------------------------------

def get_check_unix_ascii(ftp, fname):
    """
        Get a Unix/Linux ascii file from an open FTP connection and
        check that local filesize is the same as the remote file size.


        Definition
        ----------
        def get_check_unix_ascii(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        if not get_check_unix_ascii(ftp, 'test.dat'):
            print('Error in transfer')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    rsize = get_size(ftp, fname)
    get_unix_ascii(ftp, fname)
    lsize = os.stat(fname).st_size
    if rsize == lsize:
        return True
    else:
        return False

# ------------------------------------------------------------------------------------------

def get_windows_ascii(ftp, fname):
    """
        Get a Windows ascii file from an open FTP connection.


        Definition
        ----------
        def get_windows_ascii(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        get_windows_ascii(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014 - modified from
                                 http://www.java2s.com/Tutorial/Python/0420__Network/ASCIIfiledownload.htm
    """
    def writeline(data):
        fd.write(data + '\r\n') # Windows-Format
    fd = open(fname, 'wt')
    ftp.retrlines('RETR '+fname, writeline)
    fd.close()
    set_mtime(ftp, fname)

# ------------------------------------------------------------------------------------------

def get_check_windows_ascii(ftp, fname):
    """
        Get a Windows ascii file from an open FTP connection and
        check that local filesize is the same as the remote file size.


        Definition
        ----------
        def get_check_windows_ascii(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        if not get_check_windows_ascii(ftp, 'test.dat'):
            print('Error in transfer')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.
        This file is part of the JAMS Python package, distributed under the MIT License.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    rsize = get_size(ftp, fname)
    get_windows_ascii(ftp, fname)
    lsize = os.stat(fname).st_size
    if rsize == lsize:
        return True
    else:
        return False

# ------------------------------------------------------------------------------------------

def get_names(ftp, fname=None):
    """
        Get filename(s) in current directory of open FTP connection.


        Definition
        ----------
        def get_names(ftp, fname=None):


        Input
        -----
        ftp          instance of the ftplib.FTP class


        Optional input
        --------------
        fname        filename, filename globbing is possible such as '*.dat'


        Output
        ------
        Filenames in current ftp directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fls = get_names(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    fdir = []
    if fname is not None:
        ftp.dir(fname, fdir.append)
    else:
        ftp.dir(fdir.append)
    names = []
    for f in fdir:
        i = f.split()
        if i[0][0] == '-': # assure file
            names.append(' '.join(i[8:]))      # filename
    names.sort()
    return names

# ------------------------------------------------------------------------------------------

def get_names_dates(ftp, fname=None):
    """
        Get filename(s) and file modification times in current directory of open FTP connection.


        Definition
        ----------
        def get_names_dates(ftp, fname=None):


        Input
        -----
        ftp          instance of the ftplib.FTP class


        Optional input
        --------------
        fname        filename, filename globbing is possible such as '*.dat'


        Output
        ------
        Filenames, Modification Times


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fls, flsdate = get_names_dates(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
        Modified, MC, Jan 2016 - return datetime.datetime
    """
    today    = datetime.date.today()
    thisyear = str(today.year)
    fdir = []
    if fname is not None:
        ftp.dir(fname, fdir.append)
    else:
        ftp.dir(fdir.append)
    names = []
    dates = []
    for f in fdir:
        i = f.split()
        if i[0][0] == '-': # assure file
            names.append(' '.join(i[8:]))      # filename
            yr = i[7]               # year/time column
            if ':' in yr:
                hhmm = yr
                yr = thisyear
            else:
                hhmm = '00:00'
            hhmi = hhmm.split(':')
            hh, mi = [ int(hi) for hi in hhmi ]
            yyyy  = int(yr)
            dd    = int(i[6]) # day
            mm    = i[5]      # month as Jan, Feb, ...
            mm    = datetime.datetime.strptime(mm,"%b").month
            idate = datetime.date(yyyy,mm,dd)
            if idate > today: # time instead for files younger than 6 month, even in the last year
                yyyy -= 1
            idate = datetime.datetime(yyyy,mm,dd,hh,mi)
            dates.append(idate) # file date
    ii = argsort(names)
    names = [ names[i] for i in ii ]
    dates = [ dates[i] for i in ii ]
    return names, dates

# ------------------------------------------------------------------------------------------

def get_names_dates_sizes(ftp, fname=None):
    """
        Get filename(s),  file modification times, and filesizes in current directory of open FTP connection.


        Definition
        ----------
        def get_names_dates_sizes(ftp, fname=None):


        Input
        -----
        ftp          instance of the ftplib.FTP class


        Optional input
        --------------
        fname        filename, filename globbing is possible such as '*.dat'


        Output
        ------
        Filenames, Modification Times, filesizes


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fls, flsdate, flssize = get_names_dates_sizes(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014-2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
        Modified, MC, Jan 2016 - return datetime.datetime
    """
    today    = datetime.date.today()
    thisyear = str(today.year)
    fdir = []
    if fname is not None:
        ftp.dir(fname, fdir.append)
    else:
        ftp.dir(fdir.append)
    names = []
    dates = []
    sizes = []
    for f in fdir:
        i = f.split()
        if i[0][0] == '-': # assure file
            names.append(' '.join(i[8:])) # filename
            sizes.append(int(i[4]))       # file size
            yr = i[7]         # year/time column
            if ':' in yr:
                hhmm = yr
                yr = thisyear
            else:
                hhmm = '00:00'
            hhmi = hhmm.split(':')
            hh, mi = [ int(hi) for hi in hhmi ]
            yyyy  = int(yr)
            dd    = int(i[6]) # day
            mm    = i[5]      # month as Jan, Feb, ...
            mm    = datetime.datetime.strptime(mm,"%b").month
            idate = datetime.date(yyyy,mm,dd)
            if idate > today: # time instead for files younger than 6 month, even in the last year
                yyyy -= 1
            idate = datetime.datetime(yyyy,mm,dd,hh,mi)
            dates.append(idate)           # file date
    ii = argsort(names)
    names = [ names[i] for i in ii ]
    dates = [ dates[i] for i in ii ]
    sizes = [ sizes[i] for i in ii ]
    return names, dates, sizes

# ------------------------------------------------------------------------------------------

def get_names_sizes(ftp, fname=None):
    """
        Get filename(s) and filesizes in current directory of open FTP connection.


        Definition
        ----------
        def get_names_sizes(ftp, fname=None):


        Input
        -----
        ftp          instance of the ftplib.FTP class


        Optional input
        --------------
        fname        filename, filename globbing is possible such as '*.dat'


        Output
        ------
        Filenames, filesizes


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fls, flssize = get_names_sizes(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    fdir = []
    if fname is not None:
        ftp.dir(fname, fdir.append)
    else:
        ftp.dir(fdir.append)
    names = []
    sizes = []
    for f in fdir:
        i = f.split()
        if i[0][0] == '-': # assure file
            names.append(' '.join(i[8:]))      # filename
            sizes.append(int(i[4])) # file size
    ii = argsort(names)
    names = [ names[i] for i in ii ]
    sizes = [ sizes[i] for i in ii ]
    return names, sizes

# ------------------------------------------------------------------------------------------

def get_names_times(*args, **kwargs):
    """
        Wrapper for get_names_dates:
            def get_names_dates(ftp, fname=None):


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fls, flsdate = get_names_times(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    return get_names_dates(*args, **kwargs)

# ------------------------------------------------------------------------------------------

def get_names_times_sizes(*args, **kwargs):
    """
        Wrapper for get_names_dates_sizes:
            def get_names_dates_sizes(ftp, fname=None):


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fls, flsdate, flssize = get_names_times_sizes(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    return get_names_dates_sizes(*args, **kwargs)

# ------------------------------------------------------------------------------------------

def get_size(ftp, fname):
    """
        Get file size of file on open FTP connection.


        Definition
        ----------
        def get_size(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        Filesize


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        fs = get_size(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    fdir = []
    ftp.dir(fname, fdir.append)
    return int(fdir[0].split()[4])

# ------------------------------------------------------------------------------------------

def get_sizes(ftp, fname=None):
    """
        Get filesizes of files in current directory of open FTP connection.


        Definition
        ----------
        def get_sizes(ftp, fname=None):


        Input
        -----
        ftp          instance of the ftplib.FTP class


        Optional input
        --------------
        fname        filename, filename globbing is possible such as '*.dat'


        Output
        ------
        Filesizes


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        flssize = get_sizes(ftp, '*.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    fdir = []
    if fname is not None:
        ftp.dir(fname, fdir.append)
    else:
        ftp.dir(fdir.append)
    sizes = []
    for f in fdir:
        i = f.split()
        if i[0][0] == '-': # assure file
            sizes.append(int(i[-5])) # file size
    sizes.sort()
    return sizes

# ------------------------------------------------------------------------------------------

def set_mtime(ftp, fname):
    """
        Set the access and modification times of a local file
        to time of the file with the same name in the current directory of an open FTP connection.


        Definition
        ----------
        def set_mtime(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        fname in current local directory has the same access and modification time as remote file


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        set_mtime(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2014 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Dec 2014
    """
    today     = datetime.date.today()
    thisyear  = str(today.year)
    fdir = []
    ftp.dir(fname, fdir.append)
    yr = fdir[0].split()[-2]         # year/time column
    if ':' in yr:
        hhmm = yr
        yr = thisyear
    else:
        hhmm = '00:00'
    yyyy  = int(yr)
    dd    = int(fdir[0].split()[-3]) # day
    mm    = fdir[0].split()[-4]      # month as Jan, Feb, ...
    mm    = datetime.datetime.strptime(mm,"%b").month
    idate = datetime.date(yyyy,mm,dd)
    if idate > today: # time instead for files younger than 6 month, even in the last year
        yyyy -= 1
    mtime = int(time.mktime(time.strptime(str(dd)+'.'+str(mm)+'.'+str(yyyy)+' '+hhmm, "%d.%m.%Y %H:%M")))
    os.utime(fname, (mtime, mtime))

# ------------------------------------------------------------------------------------------

def put_binary(ftp, fname):
    """
        Put a binary file to an open FTP connection.


        Definition
        ----------
        def put_binary(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        put_binary(ftp, 'test.dat')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jan 2016
    """
    ftp.voidcmd("TYPE I")
    myfile = open(fname, 'rb')
    ftp.storbinary('STOR ' + fname, myfile)
    myfile.close()

# ------------------------------------------------------------------------------------------

def put_check_binary(ftp, fname):
    """
        Put a binary file to an open FTP connection and
        check that local filesize is the same as the remote file size.


        Definition
        ----------
        def put_check_binary(ftp, fname):


        Input
        -----
        ftp          instance of the ftplib.FTP class
        fname        filename


        Output
        ------
        File fname in current local directory


        Examples
        --------
        import ftplib
        import os
        ftp = ftplib.FTP("ftp.server.de")
        ftp.login("user", "password")
        ftp.cwd('ftp/directory')
        os.chdir('local/directory')
        if not put_check_binary(ftp, 'test.dat'):
            print('Error in transfer')
        ftp.close()


        License
        -------
        This file is part of the JAMS Python package, distributed under the MIT
        License. The JAMS Python package originates from the former UFZ Python library,
        Department of Computational Hydrosystems, Helmholtz Centre for Environmental
        Research - UFZ, Leipzig, Germany.

        Copyright (c) 2016 Matthias Cuntz - mc (at) macu (dot) de

        Permission is hereby granted, free of charge, to any person obtaining a copy
        of this software and associated documentation files (the "Software"), to deal
        in the Software without restriction, including without limitation the rights
        to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
        copies of the Software, and to permit persons to whom the Software is
        furnished to do so, subject to the following conditions:

        The above copyright notice and this permission notice shall be included in all
        copies or substantial portions of the Software.

        THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
        IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
        FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
        AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
        LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
        OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
        SOFTWARE.


        History
        -------
        Written,  MC, Jan 2016
    """
    lsize = os.stat(fname).st_size
    put_binary(ftp, fname)
    rsize = get_size(ftp, fname)
    if rsize == lsize:
        return True
    else:
        return False


if __name__ == '__main__':
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



#!/usr/bin/env python
"""
This is the unittest for the General Functions module.

python -m unittest -v tests/test_general_functions.py
python -m pytest --cov=pyjams --cov-report term-missing -v tests/test_general_functions.py

"""
import unittest


class TestGeneralFunctions(unittest.TestCase):
    """
    Tests for functions/general_functions.py

    """

    def test_general_functions(self):
        import numpy as np
        import pandas as pd
        from pyjams.functions import curvature
        from pyjams.functions import dlogistic_offset, d2logistic_offset
        from pyjams.functions import dlogistic_offset_p, d2logistic_offset_p

        self.assertEqual(np.around(curvature(1., dlogistic_offset,
                                             d2logistic_offset,
                                             1., 2., 2., 1.), 4), 0.2998)
        self.assertEqual(np.around(curvature(1., dlogistic_offset_p,
                                             d2logistic_offset_p,
                                             [1., 2., 2., 1.]), 4), 0.2998)
        self.assertEqual(list(np.around(
            curvature(np.array([1., 1.]), dlogistic_offset, d2logistic_offset,
                      1., 2., 2., 1.), 4)), [0.2998, 0.2998])
        self.assertEqual(list(np.around(
            curvature(pd.Series([1., 1.]), dlogistic_offset, d2logistic_offset,
                      1., 2., 2., 1.), 4)), [0.2998, 0.2998])
        cc = curvature(pd.DataFrame([[1., 1.], [1., 1.]]),
                       dlogistic_offset, d2logistic_offset,
                       1., 2., 2., 1.)
        self.assertEqual(list(np.around(cc.to_numpy().flatten(), 4)),
                         [0.2998, 0.2998, 0.2998, 0.2998])


if __name__ == "__main__":
    unittest.main()
